/*
 * My own private optimized gluBuild2DMipmaps() function.
 *
 * Stephane Rehel, October 24 1998.
 */

/* stuff from src-glu/mipmap.c
 * i have removed gluScaleImage() function, it should be in another
 * scale.c file
 */

#ifdef PC_HEADER
#include "all.h"
#else
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "gluP.h"
#endif


/*
 * Find the value nearest to n which is also a power of two.
 */
static GLint round2( GLint n )
{
   GLint m;

   for (m=1; m<n; m*=2)
     ;

   /* m>=n */
   if (m-n <= n-m/2) {
      return m;
   }
   else {
      return m/2;
   }
}


/*
 * Given an pixel format and datatype, return the number of bytes to
 * store one pixel.
 */
static GLint bytes_per_pixel( GLenum format, GLenum type )
{
   GLint n, m;

   switch (format) {
      case GL_COLOR_INDEX:
      case GL_STENCIL_INDEX:
      case GL_DEPTH_COMPONENT:
      case GL_RED:
      case GL_GREEN:
      case GL_BLUE:
      case GL_ALPHA:
      case GL_LUMINANCE:
         n = 1;
         break;
      case GL_LUMINANCE_ALPHA:
         n = 2;
         break;
      case GL_RGB:
         n = 3;
         break;
      case GL_RGBA:
         n = 4;
         break;
      default:
         n = 0;
   }

   switch (type) {
      case GL_UNSIGNED_BYTE:    m = sizeof(GLubyte);    break;
      case GL_BYTE:             m = sizeof(GLbyte);     break;
      case GL_BITMAP:           m = 1;                  break;
      case GL_UNSIGNED_SHORT:   m = sizeof(GLushort);   break;
      case GL_SHORT:            m = sizeof(GLshort);    break;
      case GL_UNSIGNED_INT:     m = sizeof(GLuint);     break;
      case GL_INT:              m = sizeof(GLint);      break;
      case GL_FLOAT:            m = sizeof(GLfloat);    break;
      default:                  m = 0;
   }

   return n * m;
}


/**************************************************************************
 *
 * Here it really begins.
 *
 **************************************************************************/

/* components=3, format = GL_RGBA, type = UNSIGNED_BYTE */
static
void gluBuild2DMipmaps_RGB_RGBA_UNSIGNED_BYTE( GLenum target,
                                               GLint width, GLint height,
                                               GLvoid* data,
                                               GLboolean in_place )
{
  #define _GL_FORMAT GL_RGBA
  #define COMPS      4
  #define _GL_TYPE   GL_UNSIGNED_BYTE
  #define TYPE       unsigned char
  #define TTYPE      int

  #include "mipmap_fct.h"

  #undef TTYPE
  #undef TYPE
  #undef _GL_FORMAT
  #undef COMPS
  #undef _GL_TYPE
}

/* components=3, format = GL_RGB, type = UNSIGNED_BYTE */
static
void gluBuild2DMipmaps_RGB_RGB_UNSIGNED_BYTE( GLenum target,
                                              GLint width, GLint height,
                                              GLvoid* data,
                                              GLboolean in_place )
{
  #define _GL_FORMAT GL_RGB
  #define COMPS      3
  #define _GL_TYPE   GL_UNSIGNED_BYTE
  #define TYPE       unsigned char
  #define TTYPE      int

  #include "mipmap_fct.h"

  #undef TTYPE
  #undef TYPE
  #undef _GL_FORMAT
  #undef COMPS
  #undef _GL_TYPE
}

/* components=3, format = GL_RGBA, type = GL_FLOAT */
static
void gluBuild2DMipmaps_RGB_RGBA_FLOAT( GLenum target,
                                       GLint width, GLint height,
                                       GLvoid* data,
                                       GLboolean in_place )
{
  #define _GL_FORMAT GL_RGBA
  #define COMPS      4
  #define _GL_TYPE   GL_FLOAT
  #define TYPE       GLfloat
  #define TTYPE      GLfloat

  #include "mipmap_fct.h"

  #undef TTYPE
  #undef TYPE
  #undef _GL_FORMAT
  #undef COMPS
  #undef _GL_TYPE
}

/* components=3, format = GL_RGB, type = GL_FLOAT */
static
void gluBuild2DMipmaps_RGB_RGB_FLOAT( GLenum target,
                                      GLint width, GLint height,
                                      GLvoid* data,
                                      GLboolean in_place )
{
  #define _GL_FORMAT GL_RGB
  #define COMPS      3
  #define _GL_TYPE   GL_FLOAT
  #define TYPE       GLfloat
  #define TTYPE      GLfloat

  #include "mipmap_fct.h"

  #undef TTYPE
  #undef TYPE
  #undef _GL_FORMAT
  #undef COMPS
  #undef _GL_TYPE
}

GLint APIENTRY my_gluBuild2DMipmaps( GLenum target, GLint components,
                                     GLint width, GLint height, GLenum format,
                                     GLenum type, const void *data )
{
   GLint w, h, maxsize;
   void *image, *newimage;
   GLint neww, newh, level, bpp;
   int error;
   GLboolean simple_pack;
   GLboolean RGB_comp;

   if (width < 1 || height < 1)
      return GLU_INVALID_VALUE;

   glGetIntegerv( GL_MAX_TEXTURE_SIZE, &maxsize );

   w = round2( width );
   if (w>maxsize) {
      w = maxsize;
   }
   h = round2( height );
   if (h>maxsize) {
      h = maxsize;
   }

   bpp = bytes_per_pixel( format, type );
   if (bpp==0) {
      /* probably a bad format or type enum */
      return GLU_INVALID_ENUM;
   }

   if (w!=width || h!=height) {
      /* must rescale image to get "top" mipmap texture image */
      image = malloc( (w+4) * h * bpp );
      if (!image) {
         return GLU_OUT_OF_MEMORY;
      }
      error = gluScaleImage( format, width, height, type, data,
                             w, h, type, image );
      if (error) {
         return error;
      }
   }
   else {
      image = (void *) data;
   }

   /* Brian: is it necessary ?? */
   {
   GLint unpackrowlength, unpackalignment, unpackskiprows, unpackskippixels;
   GLint packrowlength, packalignment, packskiprows, packskippixels;
   glGetIntegerv( GL_UNPACK_ROW_LENGTH, &unpackrowlength );
   glGetIntegerv( GL_UNPACK_ALIGNMENT, &unpackalignment );
   glGetIntegerv( GL_UNPACK_SKIP_ROWS, &unpackskiprows );
   glGetIntegerv( GL_UNPACK_SKIP_PIXELS, &unpackskippixels );
   glGetIntegerv( GL_PACK_ROW_LENGTH, &packrowlength );
   glGetIntegerv( GL_PACK_ALIGNMENT, &packalignment );
   glGetIntegerv( GL_PACK_SKIP_ROWS, &packskiprows );
   glGetIntegerv( GL_PACK_SKIP_PIXELS, &packskippixels );
   simple_pack= (unpackrowlength==0 && unpackalignment ==4 &&
                 unpackskiprows ==0 && unpackskippixels==0 &&
                   packrowlength==0 &&   packalignment ==4 &&
                   packskiprows ==0 &&   packskippixels==0);
   }

   /* Brian: should we optimize it for GL_RGBA too?? */
   RGB_comp= (components==3 || components==GL_RGB);

   if( RGB_comp && simple_pack && format == GL_RGBA && type == GL_UNSIGNED_BYTE )
     {
     gluBuild2DMipmaps_RGB_RGBA_UNSIGNED_BYTE( target,
                                               w, h, image, image!=data );
     }
    else
   if( RGB_comp && simple_pack && format == GL_RGB && type == GL_UNSIGNED_BYTE )
     {
     gluBuild2DMipmaps_RGB_RGB_UNSIGNED_BYTE( target,
                                              w, h, image, image!=data );
     }
    else
   if( RGB_comp && simple_pack && format == GL_RGBA && type == GL_FLOAT )
     {
     gluBuild2DMipmaps_RGB_RGBA_FLOAT( target,
                                       w, h, image, image!=data );
     }
    else
   if( RGB_comp && simple_pack && format == GL_RGB && type == GL_FLOAT )
     {
     gluBuild2DMipmaps_RGB_RGB_FLOAT( target,
                                      w, h, image, image!=data );
     }
    else
     {
     /* the original function */

     level = 0;
     while (1)
       {
       glTexImage2D( target, level, components, w, h, 0, format, type, image );

       if (w==1 && h==1)  break;

       neww = (w<2) ? 1 : w/2;
       newh = (h<2) ? 1 : h/2;
       newimage = malloc( (neww+4) * newh * bpp );
       if (!newimage)
          return GLU_OUT_OF_MEMORY;

       error =  gluScaleImage( format, w, h, type, image,
                               neww, newh, type, newimage );
       if (error)
          return error;

       if (image!=data)
          free( image );
       image = newimage;

       w = neww;
       h = newh;
       level++;
       }
     }

   if (image!=data)
      free( image );

   return 0;
}
