#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "GL/glut.h"

static GLuint Window = 0;
static GLfloat Angle = 0.0f;

#define USE_MIPMAP
//#define USE_GLU_MIPMAPPING

#define GL_COMP GL_RGB
#define COMP    3

#define TYPE_UNSIGNED_BYTE
//#define TYPE_UNSIGNED_SHORT
//#define TYPE_INT
//#define TYPE_FLOAT

#ifdef TYPE_UNSIGNED_BYTE
#define GL_TYPE GL_UNSIGNED_BYTE
#define TYPE unsigned char
#define BYTE_TO_TYPE(b) (b)
#endif

#ifdef TYPE_UNSIGNED_SHORT
#define GL_TYPE GL_UNSIGNED_SHORT
#define TYPE unsigned short
#define BYTE_TO_TYPE(b) ((b)*255)
#endif

#ifdef TYPE_FLOAT
#define GL_TYPE GL_FLOAT
#define TYPE float
#define BYTE_TO_TYPE(b) ((float)(b)/255.0F)
#endif

#ifdef TYPE_INT
#define GL_TYPE GL_INT
#define TYPE int
#define BYTE_TO_TYPE(b) ((b)*127)
#endif

TYPE* texture= NULL;
int width=0, height=0;

void read_image()
{
  FILE* file= fopen("image-XY.pnm","r");
  unsigned char line[10000];
  int i, j;

  if( !file )
    {
    printf("unable to open texture image file\n");
    exit(1);
    }

  fgets( line, 99, file );

  for(;;)
    {
    fgets( line, 99, file );
    if( line[0] != '#' )
      break;
    }
  sscanf( line, "%d %d\n", &width, &height );
  fgets( line, 99, file );

  printf( "this is a %dx%d texture\n", width, height );

  texture= (TYPE*) malloc( width * height * 4*sizeof(TYPE) );
  for( j= 0; j < height; ++j )
    {
    TYPE* l= texture + width*COMP*j;
    fread( line, width*3, 1, file );
    for( i= 0; i < width; ++i )
      {
      l[i*COMP+0]= (TYPE) BYTE_TO_TYPE(line[i*3+0]);
      l[i*COMP+1]= (TYPE) BYTE_TO_TYPE(line[i*3+1]);
      l[i*COMP+2]= (TYPE) BYTE_TO_TYPE(line[i*3+2]);
      l[i*COMP+3]= 0; /* to be sure */
      }
    }
  fclose(file);
}


double get_time()
{
  static struct timeval tv;
  static struct timezone tz;

  gettimeofday( &tv, &tz );

  return (double)tv.tv_sec + ((double)tv.tv_usec)*1e-6;
}

GLint my_gluBuild2DMipmaps( GLenum target, GLint components,
                            GLint width, GLint height, GLenum format,
                            GLenum type, const void *data );


void send_mipmap()
{
#ifdef USE_MIPMAP
  #ifdef USE_GLU_MIPMAPPING
    gluBuild2DMipmaps( GL_TEXTURE_2D, GL_RGB, width, height,
                       GL_COMP, GL_TYPE, texture );
  #else
    my_gluBuild2DMipmaps( GL_TEXTURE_2D, GL_RGB, width, height,
                          GL_COMP, GL_TYPE, texture );
  #endif
#else
  /* no mimapping */
  glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, width, height, 0,
                GL_COMP, GL_TYPE, texture );
#endif /* USE_MIPMAP */
}


static void draw( void )
{
   static int nn= 0;
   static double t1, t2, mm_t1, mm_t2, mm_delta= 0.0;

  if( ! nn ) t1= get_time(); /*  t1= get_time();*/

/*   glDepthFunc(GL_EQUAL);*/
   /*   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );*/
   glClear( GL_COLOR_BUFFER_BIT );

   glColor3f( 1.0, 1.0, 1.0 );

   glPushMatrix();
glScalef(2,2,2);
{
float aa= sin(Angle/180.0*3.14) *10.0;
glRotatef( aa, 0.0, 1.0, 0.0 );
glRotatef( aa, 1.0, 0.5, 0.0 );
}
/*   glTranslatef( 1.0, 0.0, 0.0 );*/
/*   glRotatef( Angle-90.0, 0.0, 1.0, 0.0 );*/

  mm_t1= get_time(); /*  mm_t1= get_time();*/

#ifdef USE_MIPMAP
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
#endif

  send_mipmap();

   mm_t2= get_time();
   mm_delta+= mm_t2 - mm_t1;

   glBegin( GL_POLYGON );
   glTexCoord2f( 0.0, 1.0 );   glVertex2f( -1.0, -1.0 );
   glTexCoord2f( 1.0, 1.0 );   glVertex2f(  1.0, -1.0 );
   glTexCoord2f( 1.0, 0.0 );   glVertex2f(  1.0,  1.0 );
   glTexCoord2f( 0.0, 0.0 );   glVertex2f( -1.0,  1.0 );

   glEnd();
   glPopMatrix();

   glutSwapBuffers();

   t2= get_time();
   ++nn;
   printf( "\r%.2f img/s (mipmap=%.3f s)",
           1.0/((t2-t1)/(double)(nn)), mm_delta/(double)(nn) );
   fflush(stdout);
/*   printf("\r%.2f img/s", 1.0/(t2-t1) ); fflush(stdout);*/
}

static void idle( void )
{
   Angle += 2.0;
   glutPostRedisplay();
}

/* change view Angle, exit upon ESC */
static void key(unsigned char k, int x, int y)
{
   switch (k) {
      case 27:
         glutDestroyWindow(Window);
         exit(0);
   }
}

/* new window size or exposure */
static void reshape( int width, int height )
{
   glViewport(0, 0, (GLint)width, (GLint)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   /*   glOrtho( -3.0, 3.0, -3.0, 3.0, -10.0, 10.0 );*/
   glFrustum( -2.0, 2.0, -2.0, 2.0, 6.0, 20.0 );
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glTranslatef( 0.0, 0.0, -8.0 );
}

static void init( void )
{
   read_image();

   glDisable( GL_DITHER );

   /* Setup texturing */
   glEnable( GL_TEXTURE_2D );
   glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );
   glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST );

   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
}

int main( int argc, char *argv[] )
{
   glutInitWindowPosition(0, 0);
   glutInitWindowSize(300, 300);
   glutInitDisplayMode( GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE );

   Window = glutCreateWindow("Texture Objects");
   if (!Window)
      exit(1);

   init();

   glutReshapeFunc( reshape );
   glutKeyboardFunc( key );
   glutIdleFunc( idle );
   glutDisplayFunc( draw );
   glutMainLoop();
   return 0;
}
